/*
 * cli.c
 * Greg Cook, 1/Feb/2022
 *
 * Base91: encodes and decodes files in base91 ASCII
 *
 * Command line interface
 */

/* 2020-08-21: add -v flag
 * 2019-11-02: initialize optind, opterr for *Go reuse
 * 2019-10-03: use non-underscore library function names
 * 2018-12-29: add -p flag
 * 2017-09-29: add -A flag
 * 2011-09-07: source and destination file in error reports
 * 2011-09-01: split into modules
 * 2011-08-31: all code in place, compiled, first output
 * 2011-08-29: all in C, still some work to do
 * 2011-08-28: replaced state machine, implemented 32-bit crc, pos
 *             context
 * 2011-08-20: the decoder compiles...
 * 2011-08-20: reworked decoder to version 2
 * 2011-08-18: finished decoder
 * 2011-08-17: getopt loop done, decode started, reached state machine
 * 2011-08-16: main tables done, getopt loop started
 * 2011-08-15: conversion to C started */

#ifdef _WIN32
#  include <io.h>
#  include <fcntl.h>
#  ifndef STDIN_FILENO
#    define STDIN_FILENO 0
#  endif
#  ifndef STDOUT_FILENO
#    define STDOUT_FILENO 1
#  endif
#endif

#include "base91.h"

const char *myname = "base91", *src = NULL, *dest = NULL;

int
main(int argc, char *argv[]) {
  /* Process command line options and execute specified function. */
  int flags = U_CRC | U_CRCD | U_POS | U_RLE;
  int c;
  char *suffix;
  size_t blocsz = 0;
  size_t base = 1;
  char *ofile = NULL;
  int col = 0;
  FILE *input;
  char *narg = "", **nargv = &narg;

  myname = argv[0];
  src = NULL; dest = NULL;

  /* reset getopt() with our extension in case user used *Go */
  optind = 0; opterr = 0;
  getopt(1, nargv, narg); /* getopt() sees end of argument 0 */
  getopt(1, nargv, narg); /* getopt() sees end of arguments */
  optind = 1; opterr = 1;

  do {
    c=getopt(argc, argv, "?ACDPRb:cdhio:puvw:");
    switch(c) {
      case '?': /* ?  get help / usage */
      case 'h': /* h  get help / usage */
      case 'u': /* u  get help / usage */
      default:
        usage();
        exit(EXIT_FAILURE);
        break;
      case 'A': /* A  write Type A stream */
        flags |= U_TYPEA;
        break;
      case 'C': /* C  do not use CRCs */
        flags &= ~(U_CRC | U_CRCD);
        break;
      case 'D': /* D  do not use CRC-32D */
        flags &= ~U_CRCD;
        break;
      case 'P': /* P  do not use positions */
        flags &= ~U_POS;
        break;
      case 'R': /* R  do not use RLE */
        flags &= ~U_RLE;
        break;
      case 'b': /* b: block size */
        blocsz = (size_t) strtoul(optarg, &suffix, 10);
        switch(*suffix) {
          case 'G':
            base = 1000000000;
            break;
          case 'M':
            base = 1000000;
            break;
          case 'k':
            base = 1000;
            break;
          case 'g':
            base = 0x40000000;
            break;
          case 'm':
            base = 0x100000;
            break;
          case 'K':
            base = 0x400;
            break;
          case 'B':
            base = 0x200;
            break;
          case 0:
            base = 1;
            break;
          default:
            error("unrecognised block size suffix");
        }
        if(blocsz > (size_t) ~0UL / base) {
          error("block size overflow");
        }
        blocsz *= base;
        break;
      case 'c': /* c  decode to standard output */
        flags |= U_STDOUT;
        break;
      case 'd': /* d  decode */
        flags |= U_DECODE;
        break;
      case 'i': /* i  integrity checks and new line for every block */
        flags |= U_INTCHK;
        break;
      case 'o': /* o: name of output file */
        ofile = optarg;
        flags |= U_OFILE;
        break;
      case 'p': /* p  pad lines with whitespace */
        flags |= U_PAD;
        break;
      case 'v': /* v  verbose */
        flags |= U_SUMMAR;
        break;
      case 'w': /* w: output wrap width */
        col = atoi(optarg);
        break;
      case -1: /* no more options, continue */
        ;
    }
  } while(c != -1);

  if(flags & U_DECODE) {
    /* Decode */
    /* stdout must be binary */
#ifdef _WIN32
    setmode(STDOUT_FILENO, O_BINARY);
#endif
    if(optind == argc) {
      /* If no arguments, decode standard input */
      src = "-";
      decode(stdin, ofile, flags);
    } else {
      /* Open each named file and decode its contents (separately) */
      for(; optind < argc; ++optind) {
        src = argv[optind];
        if(strcmp(argv[optind], "-")) {
          /* If filename is not "-", decode the file */
          if(!(input = fopen(argv[optind], "rb")))
            error("cannot open for reading");
          decode(input, ofile, flags);
          if(ferror(input))
            error("error condition on file");
          if(input != stdin)
            fclose(input);
        } else {
          /* If filename is "-", decode standard input */
          decode(stdin, ofile, flags);
        }
      }
    }
  } else {
    /* Encode */
    /* stdin must be binary */
#ifdef _WIN32
    setmode(STDIN_FILENO, O_BINARY);
#endif
    if(flags & U_STDOUT || optind == argc) {
      /* If -c or no arguments, encode standard input */
      encode(stdout, "-", ofile, flags, blocsz, col);
    } else {
      for(; optind < argc; ++optind) {
        /* Pass each filename for encoding.
         * NB encode() takes care of the "-" case */
        encode(stdout, argv[optind], ofile, flags, blocsz, col);
      }
    }
    /* Mark the end of the base91 message. */
    endenc();
  }
  exit(EXIT_SUCCESS);
}

void
error(const char *msg) {
  /* Display an error message with source, and abort */
  if(dest)
    fprintf(stderr,"%s: %s: %s: %s\n", myname, src, dest, msg);
  else if(src)
    fprintf(stderr,"%s: %s: %s\n", myname, src, msg);
  else
    fprintf(stderr,"%s: %s\n", myname, msg);
  exit(EXIT_FAILURE);
}

void
usage(void) {
  /* Print usage instructions to standard error */
  fputs("Base91: encodes and decodes files in base91 ASCII.  "
    "Level 1, version " VERSION "\n"
    "Usage:\t",stderr);
  fputs(myname, stderr);
  fputs("\n\t[-AcCdDipPRhu?] [-b BLOCKSIZE] [-o OUTFILE] "
    "[-w COLUMNS] [FILE...]\n"
    "\t-d\t\tdecode (default is to encode)\n"
    "\t-h | -u | -?\tdisplay this help\n"
    "Decoding options:\n"
    "\t-c\t\tdecode to standard output\n"
    "\t-o OUTFILE\twrite to file OUTFILE instead of name given\n"
    "\t-v\t\tprint summary of each file to standard error\n"
    "Encoding options:\n"
    "\t-A\t\temit a Type A stream (default is Type B)\n"
    "\t-c\t\tencode from standard input\n"
    "\t-C\t\tdo not add CRCs\n"
    "\t-D\t\tlimit CRCs to 32 bits\n",stderr);
  fputs("\t-i\t\tintegrity checks and new line for every block\n"
    "\t-p\t\tpad stream lines with space (to paste into OpenSSH)\n"
    "\t-P\t\tdo not add position markers\n"
    "\t-R\t\tdo not use run-length encoding\n"
    "\t-b BLOCKSIZE\t"
    "encode in blocks of BLOCKSIZE octets (default 32K)\n"
    "\t\t\tsuffixes: k = 10^3   M = 10^6\tG = 10^9\n"
    "\t\t\tB = 2^9\t  K = 2^10   m = 2^20\tg = 2^30\n"
    "\t-o OUTFILE\tgive filename as OUTFILE instead of FILE...\n"
    "\t-w COLUMNS\twrap output to COLUMNS columns (default 76)\n",
    stderr);
}

/* End of cli.c */
/* vim:tw=72:nocin:ai:et:
*/
