/*
 * decode.c
 * Greg Cook, 21/Aug/2020
 *
 * Base91: encodes and decodes files in base91 ASCII
 *
 * Decoding functions
 */

#include "base91.h"

static unsigned int getsym(FILE *input);
static unsigned int getlsm(FILE *input);
static INLINE void wrchar(unsigned int c);
static INLINE void wrshrt(unsigned int s);
static INLINE void wrlong(unsigned long l);
static void checkc(unsigned long accl);
static void checkd(unsigned long acch);
static void checkp(unsigned long accl, unsigned long acch);
static void fpru64(FILE *stream, int c,
  unsigned long hi, unsigned long lo);

static FILE *output = NULL;
static unsigned long posl = 0UL, posh = 0UL;
static unsigned long crcc = 0xFFFFFFFFUL, crcd = 0xFFFFFFFFUL;

void
decode(FILE *input, const char *ofile, int flags) {
  /* Decode a base91 message on stream input. */
  enum dstate {
    s_srch,
    s_head,
    s_data,
    s_crc,
    s_pos
  } state = s_srch;
  char buf[LNSZ], *line;
  const char *fname;
  size_t len;
  unsigned long accl = 0UL, acch = 0UL;
  unsigned int symbol, count, limit;
#ifdef MODES
  int md, um;
#endif

  do {
    /* Until an end line is found, search for a begin line */
    if(!(line = fgets(buf, LNSZ, input)))
    /* On end-of-file, break out of loop without error */
      break;
    /* Skip initial whitespace, trim off newline */
    line = buf + strspn(buf, "\t\r ");
    if((len = strlen(line)) && line[len-1] == '\n')
      line[--len] = 0;
    if(len && line[len-1] == '\r')
      line[--len] = 0;
    /* A begin line matches this:
     * /^[\t\r ]*begin-base91 [0-7]+ (.*)\r?$/
     */
    if(!strncmp(line, "begin-base91 ", (size_t) 13)
        && (len = strspn(line+13, "01234567"))
        && line[13+len] == ' ') {
      fname = line+14+len;
#ifdef MODES
      md = (int) (0777UL & strtoul(line+13, NULL, 8));
      /* Clear permission bits on the new file
       * until we say what they should be.
       */
      um = umask(0777);
#endif

      /* override output name */
      if(flags & U_OFILE)
        fname = ofile;

      /* open output */
      dest = fname;
      if(flags & U_STDOUT
          || !strcmp(fname,"-")) {
        dest = "-";
        output = stdout;
      } else {
        if(!(output = fopen(fname, "wb"))) {
          error("cannot open for writing");
        }
      }
      posh = posl = 0UL; crcd = crcc = 0xFFFFFFFFUL; limit = 0x2040U;
      state = s_head;

      /* Read symbols until an end marker is found. */
      symbol = getsym(input);
      while(symbol != 0x2048U) {
        /* Recognise a Type B stream. Can be changed to a fixed
         * two-symbol sequence if necessary (as long as interpretation
         * in either Type does not result in a frame shift).
         */
        if(symbol == 0x2058U && state == s_head && limit == 0x2040U) {
          limit = 0x2000U;
          symbol = getsym(input);
        } else if(symbol < 0x2048U || symbol >= 0x2050U) {
        /* Data word follows, validate word length and state */
          switch(state) {
            case s_head:
              error("unexpected data in header context");
              break;
            case s_crc:
              /* 65/32 only, not var/16/13 */
              if(symbol >= 0x2040U)
                error("unexpected data word length in CRC context");
              break;
            case s_pos:
              /* 65/32/16 only, not var/13 */
              if((symbol >= 0x2040U && symbol < 0x2048U)
                || symbol == 0x2058U)
                error("unexpected data word length "
                  "in position context");
              break;
            default:
              ;
          }
          /* Treat data words by length: */
          if(symbol < 0x2000U) {
            /* 65-bit word */
            accl = symbol | ((unsigned long) getlsm(input) << 13);
            symbol = getlsm(input);
            accl |= (unsigned long) (symbol & 0x3FU) << 26;
            acch = (symbol >> 6) | ((unsigned long) getlsm(input) << 7);
            symbol = getlsm(input);
            if(symbol >= 0x1000U)
              error("unexpected set flag in data word");
            acch |= (unsigned long) symbol << 20;
            switch(state) {
              case s_data:
                /* Write data word to output stream */
                wrlong(accl);
                wrlong(acch);
                break;
              case s_crc:
                /* Check CRC word against calculation
                 * then reset CRC and go to data context
                 */
                checkc(accl);
                checkd(acch);
                state = s_data;
                crcd = crcc = 0xFFFFFFFFUL;
                break;
              case s_pos:
                /* Check position word against position
                 * then go to data context
                 */
                checkp(accl, acch);
                state = s_data;
                break;
              default:
                ;
            }
            symbol = getsym(input);
          } else if(symbol < 0x2040U) {
            /* 32-bit word */
            accl = (symbol & 0x3FU) | ((unsigned long) getlsm(input) << 6)
              | ((unsigned long) getlsm(input) << 19);
            switch(state) {
              case s_data:
                /* Write data word to output stream */
                wrlong(accl);
                break;
              case s_crc:
                /* Check CRC word against calculation
                 * then reset CRC and go to data context
                 */
                checkc(accl);
                state = s_data;
                crcd = crcc = 0xFFFFFFFFUL;
                break;
              case s_pos:
                /* Check position word against position
                 * then go to data context
                 */
                checkp(accl, 0UL);
                state = s_data;
                break;
              default:
                ;
            }
            /* read next symbol and loop */
            symbol = getsym(input);
          } else if(symbol < 0x2048U) {
            /* variable-length word, 0 to 15 bits */
            accl = (symbol & 0x7U) | ((unsigned long) getlsm(input) << 3);
            if(!accl)
              /* Word 0x2040, 0x0000 undefined. */
              error("unrecognised variable-length word");
            else if((accl | 0xFFUL) == 0x1FFUL)
              /* 8 bits of data. */
              wrchar(accl & 0xFFUL);
            else if(accl != 1UL)
              /* Word 0x2041, 0x0000 means no data;
               * otherwise, this implementation doesn't deal
               * with partial octets.
               */
              error("unexpected variable word length");
            /* read next symbol and loop */
            symbol = getsym(input);
          } else if(symbol < 0x2058U) {
            /* 16-bit word */
            accl = (symbol & 0x7U) | ((unsigned long) getlsm(input) << 3);
            switch(state) {
              case s_data:
                /* Write data word to output stream */
                wrshrt(accl);
                break;
              case s_pos:
                /* Check position word against position
                 * then go to data context
                 */
                checkp(accl, 0UL);
                state = s_data;
                break;
              default:
                ;
            }
            /* read next symbol and loop */
            symbol = getsym(input);
          } else {
            /* 13-bit word */
            do {
              /* read symbols until one of 0x1200 or more is found */
              symbol = getsym(input);
              if(symbol < 0x2000U) {
                /* A run of octets, write to output stream */
                accl = symbol & 0xFF;
                accl |= (accl << 8);
                accl |= (accl << 16);
                count = counts[symbol >> 8];
                while(count >= 4) {
                  wrlong(accl);
                  count -= 4;
                }
                if(count & 2)
                  wrshrt(accl);
                if(count & 1)
                  wrchar(accl);
              } else if(symbol >= limit) {
                ;
              } else if(symbol == 0x2000U)
                /* Word 0x2000 is undefined in 13-bit context */
                error("unrecognised variable-length word");
              else if(symbol >= 0x2002U)
                /* This implementation doesn't deal with
                 * parts of octets
                 */
                error("unexpected variable word length");
            } while(symbol < 0x1200U);
            if(symbol < limit)
              /* Symbol >= limit has not been spent,
               * so loop to interpret.
               * Otherwise read next symbol and loop
               */
              symbol = getsym(input);
          }
        /* Markers */
        } else if(symbol == 0x2049U) {
          /* Data context, reset CRC */
          state = s_data;
          crcd = crcc = 0xFFFFFFFFUL;
          symbol = getsym(input);
        } else if(symbol == 0x204AU) {
          /* Position context, reset CRC if in header context */
          if(state == s_head)
            crcd = crcc = 0xFFFFFFFFUL;
          state = s_pos;
          symbol = getsym(input);
        } else if(symbol == 0x204BU) {
          /* CRC context */
          state = s_crc;
          symbol = getsym(input);
        } else if(symbol >= 0x204CU)
          error("unrecognised marker");
      }
      /* Now sitting just after the end-of-stream marker.
       * Close file.
       */
      if(output != stdout && fclose(output))
        error("error writing to file");

#ifdef MODES
      /* Set permissions from the begin line,
       * then restore the user mask
       */
      chmod(fname, md);
      umask(um);
#endif

      /* Print summary */
      if(flags & U_SUMMAR) {
        fprintf(stderr, "'%s', ",fname);
        fpru64(stderr, 0, posh, posl);
        fprintf(stderr, " bits (");
        /* Divide position by 8 and round up */
        symbol = (unsigned int) (posl & 7UL);
        posl = (posl >> 3 | posh << 29) & 0xFFFFFFFFUL;
        posh = posh >> 3 & 0xFFFFFFFFUL;
        if(symbol && !(posl = (posl + 1) & 0xFFFFFFFFUL))
          posh = (posh + 1) & 0xFFFFFFFFUL;
        fpru64(stderr, 0, posh, posl);
        fprintf(stderr, " octets)\n");
      }

      /* Make sure we are at the end of a line. */
      if((line = fgets(buf, LNSZ, input))
        && strcmp(line,"\n") && strcmp(line, "\r\n"))
        error("junk on line after end marker");
    } /* dealt with begin line */
    dest = NULL;
  } while(line && strcmp(line, "end"));
  /* look for next begin line */
  if(state == s_srch)
    error("no begin line");
  if(!line)
    error("no end line");
}

static unsigned int
getsym(FILE *input) {
  /* Read two base91 characters, skipping whitespace
   * (but not allowing a newline between them)
   * Then return the value of the combined symbol
   */

  /* Reverse lookup table of ASCII value to base91 number. */
  static const unsigned char rreper[] = {
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
    0xFF, 0x7F, 0x7F, 0xFF, 0xFF, 0x7F, 0xFF, 0xFF,
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
    0x7F, 0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46,
    0x47, 0x48, 0x49, 0x3E, 0x4A, 0xFF, 0x4B, 0x3F,
    0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B,
    0x3C, 0x3D, 0x4C, 0x4D, 0x4E, 0xFF, 0xFF, 0x4F,
    0x50, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06,
    0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E,
    0x0F, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16,
    0x17, 0x18, 0x19, 0x51, 0x52, 0x53, 0x54, 0x55,
    0x56, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F, 0x20,
    0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28,
    0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F, 0x30,
    0x31, 0x32, 0x33, 0x57, 0x58, 0x59, 0x5A, 0xFF,
  };
  int c, d;

  do {
    if((c = fgetc(input)) < 0)
      error("unexpected end of stream");
    if(c >= 0x80 || (c = rreper[c]) & 0x80)
      error("illegal character in stream");
  } while(c >= 91);
  do {
    if((d = fgetc(input)) < 0)
      error("unexpected end of stream");
    if(d == '\n')
      error("character missing from stream");
    if(d >= 0x80 || (d = rreper[d]) & 0x80)
      error("illegal character in stream");
  } while(d >= 91);
  return(91U * d + c);
}

static unsigned int
getlsm(FILE *input) {
  /* Read a symbol and ensure that it is in the correct range
   * for the trailing symbols of a data word
   */
  unsigned int sym;
  if((sym = getsym(input)) >= 0x2000U)
    error("unexpected symbol in data word");
  return(sym);
}

static INLINE void
wrchar(unsigned int c) {
  /* Write an octet to the output stream
   * and update the CRCs and position
   */
  if(fputc((int) (c & 0xFFU), output) == EOF)
    error("error writing to file");

  crcc = crc32c[(crcc ^ c) & 0xFFU] ^ (crcc >> 8);
  crcd = crc32d[(crcd ^ c) & 0xFFU] ^ (crcd >> 8);

  if((posl = (posl + 8UL) & 0xFFFFFFFFUL) < 8UL)
    ++posh;
}

static INLINE void
wrshrt(unsigned int s) {
  /* Write a little-endian 16-bit word to the output stream
   * and update the CRCs and position
   */
  unsigned long c;
  s &= 0xFFFFU;
  if(fputc((int) (s & 0xFFU), output) == EOF
    || fputc((int) (s >> 8), output) == EOF)
    error("error writing to file");

  c    = crcc ^ s;
  c    = crc32c[c & 0xFFU] ^ (c >> 8);
  crcc = crc32c[c & 0xFFU] ^ (c >> 8);

  c    = crcd ^ s;
  c    = crc32d[c & 0xFFU] ^ (c >> 8);
  crcd = crc32d[c & 0xFFU] ^ (c >> 8);

  if((posl = (posl + 16UL) & 0xFFFFFFFFUL) < 16UL)
    ++posh;
}

static INLINE void
wrlong(unsigned long l) {
  /* Write a little-endian 32-bit word to the output stream
   * and update the CRCs and position
   */
  unsigned long c;
  if(fputc((int) (l & 0xFFU), output) == EOF
   || fputc((int) ((l >> 8) & 0xFFU), output) == EOF
   || fputc((int) ((l >> 16) & 0xFFU), output) == EOF
   || fputc((int) ((l >> 24) & 0xFFU), output) == EOF)
   error("error writing to file");

  c    = crcc ^ l;
  c    = crc32c[c & 0xFFU] ^ (c >> 8);
  c    = crc32c[c & 0xFFU] ^ (c >> 8);
  c    = crc32c[c & 0xFFU] ^ (c >> 8);
  crcc = crc32c[c & 0xFFU] ^ (c >> 8);

  c    = crcd ^ l;
  c    = crc32d[c & 0xFFU] ^ (c >> 8);
  c    = crc32d[c & 0xFFU] ^ (c >> 8);
  c    = crc32d[c & 0xFFU] ^ (c >> 8);
  crcd = crc32d[c & 0xFFU] ^ (c >> 8);

  if((posl = (posl + 32UL) & 0xFFFFFFFFUL) < 32UL)
    ++posh;
}

static void
checkc(unsigned long accl) {
  /* Compare a CRC-32C word with the calculation
   */
  if(accl != (crcc ^ 0xFFFFFFFFUL)) {
    fprintf(stderr,"%s: %s: %s: CRC-32%c error: "
      "expected 0x%08lX, actual 0x%08lX\n",
      myname, src, dest, 'C', accl, crcc ^ 0xFFFFFFFFUL);
    exit(EXIT_FAILURE);
  }
}

static void
checkd(unsigned long acch) {
  /* Compare a CRC-32D word with the calculation
   */
  if(acch != (crcd ^ 0xFFFFFFFFUL)) {
    fprintf(stderr,"%s: %s: %s: CRC-32%c error: "
      "expected 0x%08lX, actual 0x%08lX\n",
      myname, src, dest, 'D', acch, crcd ^ 0xFFFFFFFFUL);
    exit(EXIT_FAILURE);
  }
}

static void
checkp(unsigned long accl, unsigned long acch) {
  /* Compare a position word with the calculation */
  if(posl != accl || posh != acch) {
    fprintf(stderr,"%s: %s: %s: position error: "
      "expected ", myname, src, dest);
    fpru64(stderr, 0, acch, accl);
    fprintf(stderr," bits, actual ");
    fpru64(stderr, 0, posh, posl);
    fprintf(stderr," bits\n");
    exit(EXIT_FAILURE);
  }
}

static void
fpru64(FILE *stream, int c, unsigned long hi, unsigned long lo) {
  /* Print decimal 64-bit unsigned integer in low 32 bits of hi and lo
   * to stream, padded with character c if non-zero.
   */
  static const unsigned long unit[] = {
    0x89E80000UL, 0xA7640000UL, 0x5D8A0000UL, 0x6FC10000UL, 0xA4C68000UL,
    0x107A4000UL, 0x4E72A000UL, 0xD4A51000UL, 0x4876E800UL, 0x540BE400UL,
    0x3B9ACA00UL, 0x05F5E100UL, 0x00989680UL, 0x000F4240UL, 0x000186A0UL,
    0x00002710UL, 0x000003E8UL, 0x00000064UL, 0x0000000AUL, 0x00000000UL,
    0x8AC72304UL, 0x0DE0B6B3UL, 0x01634578UL, 0x002386F2UL, 0x00038D7EUL,
    0x00005AF3UL, 0x00000918UL, 0x000000E8UL, 0x00000017UL, 0x00000002UL,
    0x00000000UL,
  };

  const unsigned long *uptr = unit;
  unsigned long divh = 1UL, divl;

  hi &= 0xFFFFFFFFUL;
  lo &= 0xFFFFFFFFUL;

  while((divl = *uptr++)) {
    if(divh)
      divh = uptr[19];
    if(hi < divh || (hi == divh && lo < divl)) {
      if(c)
        fputc(c, stream);
      continue;
    }
    c = '0';
    do {
      ++c;
      if(divl > lo)
        --hi;
      hi -= divh;
      lo = (lo - divl) & 0xFFFFFFFFUL;
    } while(hi > divh || (hi == divh && lo >= divl));
    fputc(c, stream);
    c = '0';
  }
  fputc('0' + (int) lo, stream);
}

/* End of decode.c */
/* vim:tw=72:nocin:ai:et:
*/
