/*
 * encode.c
 * Greg Cook, 5/Oct/2019
 *
 * Base91: encodes and decodes files in base91 ASCII
 *
 * Encoding functions
 */

#include "base91.h"

/* Fast top symbol macro. Used in several functions */
#define SYMBOLT(accu) { \
  if(cpos > cols) newln(); cpos += 2; \
  fputc(repert[(accu) - 0x1FFEU], output); \
  fputc(0x7E, output); \
  }

static int block(const unsigned char *buf, size_t len, int flags, int b13);
static void pos(void);
static void crc(int flags);
static INLINE void qword(const unsigned char *ptr);
static INLINE void dword(const unsigned char *ptr);
static INLINE void word(const unsigned char *ptr);
static INLINE void symbol(unsigned int accu);
static INLINE void newln(void);

/* The base91 character repertoire.  Encoded in hex to prevent
 * glyph translation.  The first 64 characters are the same as
 * the base64 repertoire.
 */
static const unsigned char repert[] = {
  0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48,
  0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F, 0x50,
  0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
  0x59, 0x5A, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66,
  0x67, 0x68, 0x69, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E,
  0x6F, 0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76,
  0x77, 0x78, 0x79, 0x7A, 0x30, 0x31, 0x32, 0x33,
  0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x2B, 0x2F,
  0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28,
  0x29, 0x2A, 0x2C, 0x2E, 0x3A, 0x3B, 0x3C, 0x3F,
  0x40, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F, 0x60, 0x7B,
  0x7C, 0x7D, 0x7E,
};

static FILE *output = NULL;
static unsigned long posl = 0, posh = 0;
static unsigned long crcc = 0xFFFFFFFFUL, crcd = 0xFFFFFFFFUL;
static int cpos = 0, cols = 76;
static char padch = 0;

void
encode(FILE *out, const char *fname, const char *oname, int flags,
  size_t blocsz, int col) {
  /* Base91 encode a whole file. */
  FILE *input;
  unsigned char *buf;
  size_t getsz, gotsz;
  int b13a, md;

#ifdef MODES
  struct stat st;
#endif

  getsz = blocsz ? blocsz : BUFSZ;
  if(!(buf = (unsigned char *) malloc(getsz)))
    error("cannot allocate buffer");

  src = fname;
  if(flags & U_STDOUT || !strcmp(fname, "-")) {
    /* If -c given or filename is "-", encode standard input */
    src = "-";
    input = stdin;
  } else if(!(input = fopen(fname, "rb")))
    /* If filename is not "-", encode file */
    error("cannot open for reading");

  output = out;
  cols = ((col >= ((flags & U_PAD) ? 3 : 2)) ? col - 2 : 74);
  cpos = 0;
  padch = (flags & U_PAD) ? PADCH : 0;
  posh = posl = 0; crcd = crcc = 0xFFFFFFFFUL;

  /* Print begin line giving -o option as filename, otherwise the
   * argument.  Retrieve file permissions if available, otherwise
   * give a fixed mode.
   */
#ifdef MODES
  if(stat(fname, &st)) {
    md = 0644;
  } else {
    md = ((int) st.st_mode) & 0777;
  }
#else
  md = 0644;
#endif
  fprintf(output,"begin-base91 %o %s", md,
    flags & U_OFILE ? oname : fname);
  newln();
  /* Read first block.  gotsz contains number of characters read */
  gotsz = FREAD(buf, 1, getsz, input);
  if(gotsz) {
    /* If the file is not empty, we need a data context marker.
     * If -i given, block() will provide one
     * otherwise we emit it here.
     * If block() is emitting a Type B stream, we announce Type B
     * in the header context.
     */
    if(~flags & U_TYPEA)
      SYMBOLT(0x2058U);
    if(~flags & U_INTCHK)
      SYMBOLT(0x2049U);

    /* PITFALL: the code currently assumes that no data words shall
     * appear in the header context, i.e. the CRC calculations remain
     * reset when entering the first data context or closing the stream.
     * If data words are to appear in the header context, then the CRC
     * calculations MUST be reset at the end of the header context
     * and likewise signalled to the decoder.
     */

    /* Encode the first block. */
    b13a = block(buf, gotsz, flags, 0);
    while(gotsz == getsz) {
      /* While the last block was full, read another block
       * and encode it if not empty
       */
      gotsz = FREAD(buf, 1, getsz, input);
      if(gotsz)
        b13a = block(buf, gotsz, flags, b13a);
    }
  }
  /* Emit a CRC if we want one, and block() did not emit one. */
  if((flags & (U_CRC | U_INTCHK)) == U_CRC)
    crc(flags);
  /* Emit a final position if requested. */
  if(flags & U_POS)
    pos();
  /* Emit the end-of-stream marker. */
  SYMBOLT(0x2048U);
  fputs("\n\n", output);
  if(ferror(input))
    error("error condition on file");
  if(input != stdin)
    fclose(input);
  free(buf);
}

void
endenc(void) {
  /* Print the end line, at which the decode must stop. */
  fputs("end\n", output);
}

static void
pos(void) {
  /* Mark the current position in bits. */
  unsigned char pad[8];
  SYMBOLT(0x204AU);
  /* Convert unsigned long to little endian char array */
  pad[0] = posl & 0xFF;
  pad[1] = posl >> 8 & 0xFF;
  pad[2] = posl >> 16 & 0xFF;
  pad[3] = posl >> 24 & 0xFF;
  /* Use the shortest data word that fits. */
  if(posh) {
    pad[4] = posh & 0xFF;
    pad[5] = posh >> 8 & 0xFF;
    pad[6] = posh >> 16 & 0xFF;
    pad[7] = posh >> 24 & 0xFF;
    qword(pad);
  } else if(posl >= 0x10000) {
    dword(pad);
  } else {
    word(pad);
  }
}

static void
crc(int flags) {
  /* Emit the CRC and reset. */
  unsigned char pad[8];
  SYMBOLT(0x204BU);
  /* Convert unsigned long to little endian char array */
  pad[0] = ~crcc & 0xFF;
  pad[1] = ~crcc >> 8 & 0xFF;
  pad[2] = ~crcc >> 16 & 0xFF;
  pad[3] = ~crcc >> 24 & 0xFF;
  if(flags & U_CRCD) {
    pad[4] = ~crcd & 0xFF;
    pad[5] = ~crcd >> 8 & 0xFF;
    pad[6] = ~crcd >> 16 & 0xFF;
    pad[7] = ~crcd >> 24 & 0xFF;
    qword(pad);
  } else {
    dword(pad);
  }
  crcd = crcc = 0xFFFFFFFFUL;
}

static INLINE void
qword(const unsigned char *ptr) {
  symbol((unsigned int) *ptr | (ptr[1] << 8 & 0x1F00));
  symbol((unsigned int) ptr[1] >> 5 | ptr[2] << 3 | (ptr[3] << 11 & 0x1800));
  symbol((unsigned int) ptr[3] >> 2 | (ptr[4] << 6 & 0x1FC0));
  symbol((unsigned int) ptr[4] >> 7 | ptr[5] << 1 | (ptr[6] << 9 & 0x1E00));
  symbol((unsigned int) ptr[6] >> 4 | ptr[7] << 4);
}

static INLINE void
dword(const unsigned char *ptr) {
  SYMBOLT(0x2000U + (*ptr & 0x3F));
  symbol((unsigned int) *ptr >> 6 | ptr[1] << 2 | (ptr[2] << 10 & 0x1C00));
  symbol((unsigned int) ptr[2] >> 3 | ptr[3] << 5);
}

static INLINE void
word(const unsigned char *ptr) {
  SYMBOLT(0x2050U + (*ptr & 0x7));
  symbol((unsigned int) *ptr >> 3 | ptr[1] << 5);
}

static INLINE void
symbol(unsigned int accu) {
  /* 0 <= accu <= 0x2058 */
  if(cpos > cols)
    newln();
  cpos += 2;
  fputc(repert[accu % 91], output);
  fputc(repert[accu / 91], output);
}

static INLINE void
newln(void) {
  fputc('\n', output);
  if(padch) {
    fputc(padch, output);
    cpos = 1;
  } else {
    cpos = 0;
  }
}

/* 13-bit word concatenating macros. Only used in block() */

/* these macros cancel 13-bit mode before proceeding */
#define QWORD(ptr) {B13OFF(); qword(ptr);}
#define DWORD(ptr) {B13OFF(); dword(ptr);}
/* this macro ensures then ends 13-bit mode */
#define OCTET(ptr) {B13ON(); SYMBOL0(0x1200U | *(ptr));}
/* these macros end 13-bit mode automatically */
#define WORD(ptr) {b13=0; word(ptr);}
#define CRC(flags) {b13=0; crc(flags);}
#define POS() {b13=0; pos();}
/* cancel 13-bit mode */
#define B13OFF() {if(b13 && (flags & U_TYPEA)) SYMBOLT(0x2001U); b13=0;}
/* ensure 13-bit mode, state to be declared later */
#define B13ON() if(!b13) SYMBOLT(0x2058U);
/* emit a symbol 0x1200 or more, and end 13-bit mode if ambiguous */
#define SYMBOLT0(accu) {b13=0; SYMBOLT(accu);}
#define SYMBOL0(accu) {b13=0; symbol(accu);}
/* emit a symbol less than 0x1200, and declare 13-bit mode */
#define SYMBOL1(accu) {b13=1; symbol(accu);}

static int
block(const unsigned char *buf, size_t len, int flags, int b13) {
  /* Base91 encode a block in memory. */
  size_t i, j, l;
  unsigned int k, r;
  unsigned long c, d;

  if(flags & U_INTCHK) {
    /* If -i given, blocks provide their own data context
     * either by the data context marker or by a position
     */
    if(flags & U_POS) {
      POS();
    } else {
      SYMBOLT0(0x2049U);
      crcd = crcc = 0xFFFFFFFFUL;
    }
  }

  /* Calculate the CRCs over the block. */
  if(flags & U_CRCD) {
    c = crcc;
    d = crcd;
    for(i = 0; i < len; ++i) {
      r = buf[i];
      c = crc32c[(c ^ r) & 0xFF] ^ (c >> 8);
      d = crc32d[(d ^ r) & 0xFF] ^ (d >> 8);
    }
    crcc = c;
    crcd = d;
  } else if(flags & U_CRC) {
    c = crcc;
    for(i = 0; i < len; ++i) {
      c = crc32c[(c ^ buf[i]) & 0xFF] ^ (c >> 8);
    }
    crcc = c;
  }

  /* i equals the number of octets written so far. */
  i = 0;
  while(i<len) {
    do {
      if(flags & U_RLE) {
        /* If enabled, encode as many runs as possible. */
        do {
          /* Find the length l of the current run. */
          for(l=1; l<8192 && i+l<len
            && buf[i] == buf[i+l];++l)
            {}
          /* j is the number of octets in the run still to emit. */
          j=l; r = buf[i] & 0xFF;
          /* Search the exponential part of the counts table
           * in descending order and emit the longest run
           * not exceeding j
           */
          for(k = 0x11; k && j >= 32; --k) {
          /* if the run is between 33 and 37 inclusive,
           * emit first of two continuing 13x runs
           */
            if(j < 38 && j >= 33) {
              B13ON();
              SYMBOL1(r);
              i += 13; j -= 13;
            } else if(j >= counts[k]) {
              B13ON();
              SYMBOL1(k << 8 | r);
              i += counts[k]; j -= counts[k];
            }
          }
          if(j >= 19) {
            /* If the run exceeds 18, emit a continuing 13x run */
            B13ON();
            SYMBOL1(r);
            i += 13; j -= 13;
          }
          /* Emit a terminating linear run for the rest */
          if(j >= 6) {
            B13ON();
            SYMBOL0((j + 13) << 8 | r);
            i += j;
          }
        } while(l >= 6);
      }
      /* j, k and l now dead */
      if(len-i >= 8) {
        /* If a run follows the next two octets, emit those octets */
        if(flags & U_RLE && (r = buf[i+2]) == buf[i+3]
            && r == buf[i+4] && r == buf[i+5]
            && r == buf[i+6] && r == buf[i+7]) {
          WORD(buf+i);
          i+=2;
        } else if(b13 && (~flags & U_TYPEA)) {
          /* End 13-bit mode in Type B by emitting a double word.
           * Encodes 6 more bits in the same space than in version 2.
           */
          DWORD(buf+i);
          i+=4;
        } else if(b13) {
          /* End 13-bit mode in Type A by emitting a word.
           * Encodes 3 more bits in the same space than in version 2.
           */
          WORD(buf+i);
          i+=2;
        } else {
          /* Otherwise 13-bit mode is off; emit one quad word. */
          QWORD(buf+i);
          i+=8;
        }
      }
      /* Repeat until fewer than 8 octets remain. */
    } while(len-i >= 8);
    /* Emit the trailing data words. */
    if((len-i) & 1) {
      OCTET(buf+i);
      i+=1;
    }
    if((len-i) & 2) {
      WORD(buf+i);
      i+=2;
    }
    if((len-i) & 4) {
      DWORD(buf+i);
      i+=4;
    }
  }
  /* add 8 * i to 64-bit position, with carry */
  c = posl;
  posh += ((d = ((c >> 3) + i) & 0xFFFFFFFFUL) < (unsigned long) i) ?
    8UL : (d >> 29);
  posl = (c + ((unsigned long) i << 3)) & 0xFFFFFFFFUL;

  /* Emit optional CRC, and new line, if -i given */
  if(flags & U_INTCHK) {
    if(flags & U_CRC)
      CRC(flags);
    newln();
  }
  return(b13);
}

/* End of encode.c */
/* vim:tw=72:nocin:ai:et:
*/
